/* --------------------------------------------------------------- */
/* Licensed Materials - Property of IBM                            */
/* 5724-S84                                                        */
/* (C) Copyright IBM Corp. 2007,2008, All Rights Reserved          */
/* US Government Users Restricted Rights - Use, duplication or     */
/* disclosure restricted by GSA ADP Schedule Contract with         */
/* IBM Corp.                                                       */
/* --------------------------------------------------------------- */
/* PROLOG END TAG zYx                                              */
/*
This example showes how to do FFT2D single precision power-of-2 computation.
To enable huge page on libfft.so, export FFT_USE_HUGE_PAGE=1
Numa has been turned off for 16 SPEs.  To get better performance on 16 SPEs, use numactl -i all

Note:  
computing a forward followed by a backward transform (or vice versa) will 
result in the original data multiplied by the size of the transform 
(the product of the dimensions).

*/

#include <sys/mman.h>
#include <numa.h>
#include <libfft.h>

#include <stdio.h>
#include <libspe2.h>
#include <pthread.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <fenv.h>
#include <sys/types.h>
#include <sys/time.h>         // gettimeofday
#include <errno.h>

#include <math.h>


struct timeval timev1, timev2; // gettimeofday

/* next 3 lines used for huge page size stuff */
char *mem_file = "/huge/fft_mem.bin";
char *mem_addr0 = NULL;
char *mem_addr1 = NULL;

unsigned int NP, nspus;
unsigned int NPx, NPy;
unsigned int flags;
fft_addr_t mem0, mem1;

void *A0, *A1,  *C0, *C1;
double *AD0, *AD1,  *CD0, *CD1;
float *AF0,  *AF1,  *CF0, *CF1;
fft_2d_handle_t handle1, handle2;

#define MALLOC_BIG_ARRAYS {                                   \
   fft_addr_t raw0, raw1;                                            \
   int i;                                                     \
   if (numa_flag) {                                           \
      nodemask_zero(&mask);                                   \
      nodemask_set(&mask, 0);                                 \
      numa_bind(&mask);                                       \
   }                                                          \
   raw0 = (fft_addr_t) malloc(0x6000080);                            \
   if (!raw0) {                                               \
     printf("ERROR: unable to malloc.  Exiting...\n");        \
     return(-1);                                              \
   }   \
   mem0 = raw0; 		\
   raw0 += 0x7f;                                              \
   raw0 &= ~0x7f;                                             \
   mem_addr0 = (char *) raw0;                                 \
   for (i=0; i<0x6000000; i+=0x1000) mem_addr0[i] = 0;        \
}

double trigfunc(int inx, int iny) {
   double x, y;
   x = ((double) inx) * 6.2831853071796 / ((double) NPx);
   y = ((double) iny) * 6.2831853071796 / ((double) NPy);
   return ((double) (7.0 + sin(x) + cos(2*y) + sin(3*x) + cos(5*y)));
}


int main(int argc, char *argv[]) {
   unsigned int i, j, rc;
   int fmem=0;
   unsigned int ncycles;
   unsigned int log2_nspus, log2_NP, tile_flag, inplace_flag;

   unsigned int log2_NPx, log2_NPy;
 	
   int elapsed_microsecs;
   float elapsed_time;
   unsigned int numa_flag, largepage_flag;
   double lo_real, hi_real, lo_imag, hi_imag;
   nodemask_t mask;
   int max_node = 0;
   double tol;
   float *F0ptr, *F1ptr;
   double *D0ptr, *D1ptr;

 

	 if (argc != 9) {
     fprintf(stderr, "usage: FFT2D <tile_flag> <inplace_flag> <log_size_x> <log_size_y> <ncycles> <log2_nspus> <numa_flag> <largepage_flag>\n");
     return -1;
   }  
   

	 i = 0;
   tile_flag = atoi(argv[++i]);
   inplace_flag = atoi(argv[++i]);
   log2_NPx = atoi(argv[++i]);
   log2_NPy = atoi(argv[++i]);
   log2_NP = log2_NPx;

   ncycles = atoi(argv[++i]);
   log2_nspus = atoi(argv[++i]);
   numa_flag = atoi(argv[++i]);
   largepage_flag = atoi(argv[++i]);
 
 	 //the number of SPUs varies between 4 to 16, then log2_nspus is from 2 to 4
   if (log2_nspus < 2 || log2_nspus > 4) {
      fprintf(stderr, "please specify log2_nspus between 2 and 4\n");
      return -1;
   }
   
 
	 // the size of each dimemsion is from 32 to 2048

   if (log2_NPx < (5) || log2_NPx > (11)
   	||log2_NPy < (5) || log2_NPy > (11)) {
      fprintf(stderr, "size between 32 and 2048, please.\n"); fflush(stderr);
      return -1;
   }
   if(log2_nspus == 4)
   		numa_flag = 0;
   
   //get the number of numactl node.
   if (numa_flag) {
      if (numa_available() < 0) {
         fprintf(stderr, "your system does not support NUMA.  Turning off numa_flag...\n");
         numa_flag = 0;
      }
      else {
         max_node = numa_max_node();
          //fprintf(stderr, "NUMA reports max node is %d\n", max_node);
      }
   }
   NPx = 1 << log2_NPx;
   NPy = 1 << log2_NPy;
   NP = NPx;

   nspus = 1 << log2_nspus;

	 //allocate memory for input and output buffer
   if (largepage_flag) {
      int memsize;
      memsize = 0x6000000; 
      /* Create a large contiguous memory buffer by allocating a large
       *    * page (or more). Large page memory will also reduce the TLB thrashing.
       *       */
      if ((fmem = open (mem_file, O_CREAT | O_RDWR, 0755)) == -1) {
         printf("WARNING: unable to open file %s (errno=%d). Using malloc heap.\n", mem_file, errno);

         MALLOC_BIG_ARRAYS
      } 
      else {
         /* Delete file so that huge pages will get freed on program termination.
          */
         remove(mem_file);
     
         if (numa_flag) {
            nodemask_zero(&mask);
            nodemask_set(&mask, 0);
            numa_bind(&mask);
         }
         mem_addr0 = (char *) mmap (0, memsize, PROT_READ | PROT_WRITE, MAP_PRIVATE, fmem, 0);
   
         /* touch each of these large pages */
         if (mem_addr0 != MAP_FAILED) {
            mem_addr0[0x0800000] = (char) 0;
            mem_addr0[0x1800000] = (char) 0;
            mem_addr0[0x2800000] = (char) 0;
            mem_addr0[0x3800000] = (char) 0;
            mem_addr0[0x4800000] = (char) 0;
            mem_addr0[0x5800000] = (char) 0;

         }
         
   
         if (mem_addr0 == MAP_FAILED) {
            printf("ERROR: unable to mmap file %s (errno=%d). Using malloc heap.\n", mem_file, errno);
            close (fmem);
            MALLOC_BIG_ARRAYS
         }
      }
   }
   else MALLOC_BIG_ARRAYS
 		
   A0 = (void *) mem_addr0;
   C0 = (void *) &mem_addr0[0x3000000];
   


   AF0 = (float *) A0;
   AD0 = (double *) A0;
   CF0 = (float *) C0;
   CD0 = (double *) C0;
  
 
   //fprintf(stderr, "loading big array Data\n"); fflush(stderr);
   //fflush(stderr);

   {
      if (tile_flag) {
         for (i=0; i<NPy; ++i) for (j=0; j<NPx; ++j) {         	 
            CF0[2*(NPx*i+j)  ] = (float) trigfunc(i, j);;
            CF0[2*(NPx*i+j)+1] = 0.0f;
         }
         //transfer the format from RIRIRIRI to RRRRIIII
         fft_2d_sqp2_transform_format((void *) CF0, (void *) AF0, (size_t) NPx,(size_t) NPy, 0);
      }
      else {
         for (i=0; i<NPy; ++i) for (j=0; j<NPx; ++j) {         	
            AF0[2*(NPx*i+j)  ] = (float) trigfunc(i, j);
            AF0[2*(NPx*i+j)+1] = 0.0f;
         }
      }
   }
   //fprintf(stderr, "done loading!\n"); fflush(stderr);

	 //bind the memory to node 0
   if (numa_flag) {
      nodemask_zero(&mask);
      nodemask_set(&mask, 0);
      numa_bind(&mask);
   }

   flags = FFT_TYPE_C2C;
   flags |= (tile_flag ? FFT_TILED : 0);
   //initialize the handle
   rc = fft_2d_sp_initialize(&handle1,  1<<(log2_nspus));
   
   if (rc) { fprintf(stderr, "rc = %d\n", rc); fflush(stderr); }
   

   sleep(1);
   
   //do a forward transform and an inverse one immediately 

   for (i=0; i<2; ++i) {
      void *src, *dst;
      src = (( (i&1)) && (!inplace_flag)) ? ((void *) C0) : ((void *) A0);
      dst = ((!(i&1)) && (!inplace_flag)) ? ((void *) C0) : ((void *) A0);
      rc = fft_2d_sp_perform(handle1, NPx, NPy, &src, &dst, (unsigned int) i, flags);

   }
   if (rc) exit(0);
 

   if (tile_flag) {
			{
         fft_2d_sqp2_transform_format((void *) AF0, (void *) CF0, (size_t) NPx, (size_t) NPy, FFT_TILE_TO_RECTANGULAR);
         F0ptr = CF0;
      
      }
   }
   else {
			{
         F0ptr = AF0;

      }
   }

   hi_real = -100.0;
   lo_real =  100.0;
   hi_imag = -100.0;
   lo_imag =  100.0;

	 //calculate the error range 
   {
      tol = 3.0e-4;
      for (i=0; i<NPy; ++i) for (j=0; j<NPx; ++j) {
         double x;
         x = trigfunc(i, j);  
         //computing a forward followed by a backward transform (or vice versa) will 
				 //result in the original data multiplied by the size of the transform 
				 //(the product of the dimensions).     
         F0ptr[2*(NPx*i+j)] = F0ptr[2*(NPx*i+j)]/NPx/NPy;  
         F0ptr[2*(NPx*i+j) + 1] = F0ptr[2*(NPx*i+j) + 1]/NPx/NPy;               
         if (F0ptr[2*(NPx*i+j)] - x > hi_real) hi_real = F0ptr[2*(NPx*i+j)] - x;
         if (F0ptr[2*(NPx*i+j)] - x < lo_real) lo_real = F0ptr[2*(NPx*i+j)] - x;
         if (F0ptr[2*(NPx*i+j) + 1] > hi_imag) hi_imag = F0ptr[2*(NPx*i+j) + 1];
         if (F0ptr[2*(NPx*i+j) + 1] < lo_imag) lo_imag = F0ptr[2*(NPx*i+j) + 1];
      }
   }
   fprintf(stderr, "real err*r range = %e %e\n", lo_real, hi_real);
   fprintf(stderr, "imag err*r range = %e %e\n", lo_imag, hi_imag);
   if (lo_real < -tol || hi_real > tol || lo_imag < -tol || hi_imag > tol) {
      fprintf(stderr, "ERROR: range of error values too large...\n");
      fflush(stderr);
      //return -1;
   }
 
   //printf("now executing timing runs...\n");

   gettimeofday(&timev1, NULL);
      /* monitor timing runs */
 
 	 //run n*2 cycles to monitor timing runs
   for (i=0; i<2*ncycles; ++i) {
      void *src, *dst;
      src = (( (i&1)) && (!inplace_flag)) ? ((void *) C0) : ((void *) A0);
      dst = ((!(i&1)) && (!inplace_flag)) ? ((void *) C0) : ((void *) A0);
      rc = fft_2d_sp_perform(handle1, NPx, NPy, &src, &dst, (unsigned int) i&1, flags);
   
   }

   gettimeofday(&timev2, NULL);
   //printf("timing runs completed.\n");

	 //terminate the handle and recycle the resource
   rc = fft_2d_sp_terminate(handle1);
   if (rc) exit(0);
 
   elapsed_microsecs =
        (int)(timev2.tv_sec  - timev1.tv_sec) * 1000000 +
        (int)(timev2.tv_usec - timev1.tv_usec);
 
   elapsed_time = (float) elapsed_microsecs * 0.000001;
 
   //printf("elapsed time is %f seconds.\n", elapsed_time);
   //printf("time per FFT is %f seconds.\n", elapsed_time / (2*ncycles));
   //printf("FLOPS = %f\n", 2.0 * 5.0 * NP * NP * log2_NP * (4*ncycles));
   //printf("FLOPS/spu = %f\n", 2.0 * 5.0 * NP * NP * log2_NP * (4*ncycles) / nspus);
   printf("%d SPUs. standard '5*N*logN' GFLOPS/sec = %f\n", nspus, 0.000000001 * 5.0 * NPx * NPy * (log2_NPx + log2_NPy) * (2*ncycles) / elapsed_time);
   //printf("gigabytes/sec = %f\n", 0.000000001 * 2.0 * 2.0 * 8.0 * NP * NP * (4*ncycles) / elapsed_time);
 

   //printf("Now checking results of all the cycles...\n"); fflush(stdout);

   if (tile_flag) {
      {
      	 //transfer the format from RRRRIIII to RIRIRIRI
         fft_2d_sqp2_transform_format((void *) AF0, (void *) CF0, (size_t) NPx, (size_t) NPy, FFT_TILE_TO_RECTANGULAR);
         F0ptr = CF0;
       }
   }
   else {
      {
         F0ptr = AF0;
      }
   }
 

   free((void *)mem0);
   if (largepage_flag){
   	munmap((void *)fmem, 0x6000000);
   	close (fmem);
   }

   return 0;
}
