/* --------------------------------------------------------------- */
/* Licensed Materials - Property of IBM                            */
/* 5724-S84                                                        */
/* (C) Copyright IBM Corp. 2008       All Rights Reserved          */
/* US Government Users Restricted Rights - Use, duplication or     */
/* disclosure restricted by GSA ADP Schedule Contract with         */
/* IBM Corp.                                                       */
/* --------------------------------------------------------------- */
/* PROLOG END TAG zYx                                              */
/*
To enable huge page on libfft.so, export FFT_USE_HUGE_PAGE=1
Numa has been turned off for 16 SPEs.  To get better performance on 16 SPEs, use numactl -i all

Note:
computing an r2c followed by a c2r transform (or vice versa) will result in the original 
data multiplied by the size of the transform (the product of the logical dimensions). 
An r2c transform produces the same output (size NPx*NPy with redundancy) as a fft forward transform of the same input, 
and a c2r transform is correspondingly equivalent to fft backward tranform.
*/
#include <sys/mman.h>
#include <numa.h>
#include <libfft.h>

#include <stdio.h>
#include <libspe2.h>
#include <pthread.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <fenv.h>
#include <sys/types.h>
#include <sys/time.h>         // gettimeofday
#include <errno.h>

#include <math.h>


struct timeval timev1, timev2; // gettimeofday

/* next two lines used for huge page size stuff */
char *mem_file = "/huge/fft_mem.bin";
char *mem_addr0 = NULL;
char *mem_addr1 = NULL;

int NP, nspus;
int NPx, NPy;
int mem0, mem1;

void *A0, *A1,  *C0, *C1;
double *AD0, *AD1,  *CD0, *CD1;
float *AF0,  *AF1,  *CF0, *CF1;

#define MALLOC_BIG_ARRAYS {                                   \
   fft_addr_t raw0, raw1;                                            \
   int i;                                                     \
   if (numa_flag) {                                           \
      nodemask_zero(&mask);                                   \
      nodemask_set(&mask, 0);                                 \
      numa_bind(&mask);                                       \
   }                                                          \
   raw0 = (fft_addr_t) malloc(0x6000080);                            \
   if (!raw0) {                                               \
     printf("ERROR: unable to malloc.  Exiting...\n");        \
     return(-1);                                              \
   }   \
   mem0 = raw0; 		\
   raw0 += 0x7f;                                              \
   raw0 &= ~0x7f;                                             \
   mem_addr0 = (char *) raw0;                                 \
   for (i=0; i<0x6000000; i+=0x1000) mem_addr0[i] = 0;        \
}

double trigfunc(int inx, int iny) {
   double x, y;
   x = ((double) inx) * 6.2831853071796 / ((double) NPx);
   y = ((double) iny) * 6.2831853071796 / ((double) NPy);
   return ((double) (7.0 + sin(x) + cos(2*y) + sin(3*x) + cos(5*y)));
}

int main(int argc, char *argv[]) {
   int i, j, rc;
   int fmem=0;
   int ncycles=0;
   unsigned int log2_nspus, log2_NP, DPflag=0, tile_flag, inplace_flag, flags;

   unsigned int log2_NPx, log2_NPy;
 	
   int elapsed_microsecs;
   float elapsed_time;
   unsigned int numa_flag, largepage_flag;
   double lo_real, hi_real, lo_imag, hi_imag;
   nodemask_t mask;
   int max_node = 0;
   double tol;
   float *F0ptr, *F1ptr;
   double *D0ptr, *D1ptr;
   fft_2d_handle_t handle1, handle2;
 

	 if (argc != 8) {
     fprintf(stderr, "usage: FFT2D  <tile_flag> <inplace_flag> <log_size_x> <log_size_y> <log2_nspus> <numa_flag> <largepage_flag>\n");
     return -1;
   }  
   

	 i = 0;
   tile_flag = atoi(argv[++i]);
   inplace_flag = atoi(argv[++i]);
   log2_NPx = atoi(argv[++i]);
   log2_NPy = atoi(argv[++i]);
   log2_NP = log2_NPx;

   log2_nspus = atoi(argv[++i]);
   numa_flag = atoi(argv[++i]);
   largepage_flag = atoi(argv[++i]);
 
   if (log2_nspus < 2 || log2_nspus > 4) {
      fprintf(stderr, "please specify log2_nspus between 2 and 4\n");
      return -1;
   }
   
   if (log2_nspus == 4)
   		numa_flag = 0; 
 	
   if (numa_flag) {
      if (numa_available() < 0) {
         fprintf(stderr, "your system does not support NUMA.  Turning off numa_flag...\n");
         numa_flag = 0;
      }
      else {
         max_node = numa_max_node();
      }
   }

   if (log2_NPx < (5) || log2_NPx > (11)
   	||log2_NPy < (5) || log2_NPy > (11)) {
      fprintf(stderr, "size between 32 and 2048, please.\n"); fflush(stderr);
      return -1;
   }
   NPx = 1 << log2_NPx;
   NPy = 1 << log2_NPy;
   NP = NPx;

   nspus = 1 << log2_nspus;

   if (largepage_flag) {
      int memsize;
      memsize = 0x6000000; 
      /* Create a large contiguous memory buffer by allocating a large
       *    * page (or more). Large page memory will also reduce the TLB thrashing.
       *       */
      if ((fmem = open (mem_file, O_CREAT | O_RDWR, 0755)) == -1) {
         printf("WARNING: unable to open file %s (errno=%d). Using malloc heap.\n", mem_file, errno);

         MALLOC_BIG_ARRAYS
      } 
      else {
         /* Delete file so that huge pages will get freed on program termination.
          */
         remove(mem_file);
     
         if (numa_flag) {
            nodemask_zero(&mask);
            nodemask_set(&mask, 0);
            numa_bind(&mask);
         }
         mem_addr0 = (char *) mmap (0, memsize, PROT_READ | PROT_WRITE, MAP_PRIVATE, fmem, 0);
   
         /* touch each of these large pages */
         if (mem_addr0 != MAP_FAILED) {
            mem_addr0[0x0800000] = (char) 0;
            mem_addr0[0x1800000] = (char) 0;
            mem_addr0[0x2800000] = (char) 0;
            mem_addr0[0x3800000] = (char) 0;
            mem_addr0[0x4800000] = (char) 0;
            mem_addr0[0x5800000] = (char) 0;

         }
   
         if (mem_addr0 == MAP_FAILED) {
            printf("ERROR: unable to mmap file %s (errno=%d). Using malloc heap.\n", mem_file, errno);
            close (fmem);
            MALLOC_BIG_ARRAYS
         }
      }
   }
   else MALLOC_BIG_ARRAYS
 		
   A0 = (void *) mem_addr0;
   C0 = (void *) &mem_addr0[0x3000000];



   AF0 = (float *) A0;
   AD0 = (double *) A0;
   CF0 = (float *) C0;
   CD0 = (double *) C0;
 

   if (tile_flag ){
		  for (i=0; i<NPy; ++i) for (j=0; j<NPx; ++j) {        	
		      CF0[(NPx*i+j)] = (float) trigfunc(i, j);	          
		   }
	  //transfer the format from RRRRIIII to RIRIRIRI, the X dimemsion should be half of NPx
	 	fft_2d_sqp2_transform_format((void *) CF0, (void *) AF0, (size_t) NPx/2,(size_t) NPy, 0);
	}
	else	{
		for (i=0; i<NPy; ++i) for (j=0; j<NPx; ++j) {        	
      			AF0[(NPx*i+j)] = (float) trigfunc(i, j);	          
   		}
	}

   flags = 0;
   flags |= (tile_flag ? FFT_TILED : 0);
   	
   rc = fft_2d_sp_initialize(&handle1,  1<<(log2_nspus));

   if (rc) { fprintf(stderr, "rc = %d\n", rc); fflush(stderr); }




   sleep(1);
   

   
   for (i=0; i<1; ++i) {
      void *src, *dst;
	src = A0;
	dst = !inplace_flag?C0:A0;
	   if (numa_flag) {
      nodemask_zero(&mask);
      nodemask_set(&mask, 0);
      numa_bind(&mask);
   }
      rc = fft_2d_sp_perform(handle1, NPx, NPy,&src, &dst,  0, FFT_TYPE_R2C|flags);
   if (numa_flag) {
      nodemask_zero(&mask);
      nodemask_set(&mask, max_node);
      numa_bind(&mask);
   }
   /*rc = fft_2d_sp_terminate(handle1);
   if (rc) exit(0);
	printf("done r2c\n");
    rc = fft_2d_sp_initialize(&handle1,  1<<(log2_nspus));
*/
    if (rc) { fprintf(stderr, "rc = %d\n", rc); fflush(stderr); }
	src = dst;
      dst = A0;
	rc = fft_2d_sp_perform(handle1, NPx, NPy,&src, &dst,  1, FFT_TYPE_C2R|flags);
   }
   if (rc) exit(0);
   rc = fft_2d_sp_terminate(handle1);
   if (rc) exit(0);
 
   //printf("Now checking results of one cycle...\n"); fflush(stdout);
	{
         F0ptr = AF0;

      }

   hi_real = -100.0;
   lo_real =  100.0;
   hi_imag = -100.0;
   lo_imag =  100.0;

		
   {
    tol = 3.0e-4;
    for (i=0; i<NPy; ++i) for (j=0; j<NPx; ++j) {
       double x;
       x = trigfunc(i, j);     
       F0ptr[(NPx*i+j)] = F0ptr[(NPx*i+j)]/NPx/NPy;                   
       if (F0ptr[(NPx*i+j)] - x > hi_real) hi_real = F0ptr[(NPx*i+j)] - x;
       if (F0ptr[(NPx*i+j)] - x < lo_real) lo_real = F0ptr[(NPx*i+j)] - x;
       
    }
   }
   fprintf(stderr, "real err*r range = %e %e\n", lo_real, hi_real);
   if (lo_real < -tol || hi_real > tol ) {
      fprintf(stderr, "ERROR: range of error values too large...\n");
      fflush(stderr);
      return -1;
   }
   
   free((void *)mem0);
   if (largepage_flag){
   	munmap((void *)fmem, 0x6000000);
   	close (fmem);
   }
 
   return 0;
}

